import keras as ks
from keras import ops
import keras.saving


@ks.saving.register_keras_serializable(package='kgcnn', name='shifted_softplus')
def shifted_softplus(x):
    r"""Shifted soft-plus activation function.

    Args:
        x (Tensor): Single values to apply activation with ks functions.

    Returns:
        Tensor: Output tensor computed as :math:`\log(e^{x}+1) - \log(2)`.
    """
    return ks.activations.softplus(x) - ops.log(ops.convert_to_tensor(2.0, dtype=x.dtype))


@ks.saving.register_keras_serializable(package='kgcnn', name='softplus2')
def softplus2(x):
    r"""Soft-plus function that is :math:`0` at :math:`x=0` , the implementation aims at avoiding overflow
    :math:`\log(e^{x}+1) - \log(2)` .

    Args:
        x (Tensor): Single values to apply activation with ks functions.

    Returns:
         Tensor: Output tensor computed as :math:`\log(e^{x}+1) - \log(2)`.
    """
    return ks.activations.relu(x) + ops.log(0.5 * ops.exp(-ops.abs(x)) + 0.5)


@ks.saving.register_keras_serializable(package='kgcnn', name='leaky_softplus')
def leaky_softplus(x, alpha: float = 0.05):
    r"""Leaky softplus activation function similar to :obj:`tf.nn.leaky_relu` but smooth.

    .. warning::

        The leak parameter can not be changed if 'kgcnn>leaky_softplus' is passed as activation function to a layer.
        Use :obj:`kgcnn.layers.activ` activation layers instead.

    Args:
        x (Tensor): Single values to apply activation with ks functions.
        alpha (float): Leak parameter. Default is 0.05.

    Returns:
         Tensor: Output tensor.
    """
    return ks.activations.softplus(x) * (1 - alpha) + alpha * x


@ks.saving.register_keras_serializable(package='kgcnn', name='leaky_relu2')
def leaky_relu2(x, alpha: float = 0.05):
    r"""Leaky RELU activation function.

    .. warning::

        The leak parameter can not be changed if 'kgcnn>leaky_relu2' is passed as activation function to a layer.
        Use :obj:`kgcnn.layers.activ` activation layers instead.

    Args:
        x (Tensor): Single values to apply activation with ks functions.
        alpha (float): Leak parameter. Default is 0.05.

    Returns:
         Tensor: Output tensor of activations.
    """
    return ks.activations.leaky_relu(x, negative_slope=alpha)


@ks.saving.register_keras_serializable(package='kgcnn', name='swish2')
def swish2(x):
    """Swish activation function."""
    return ks.activations.silu(x)



import numpy as np
from typing import Union


def coulomb_matrix_to_inverse_distance_proton(coulomb_mat: np.ndarray,
                                              unit_conversion: float = 1.0, exponent: float = 2.4):
    r"""Convert a Coulomb matrix back to inverse distancematrix plus atomic number.

    Args:
        coulomb_mat (np.ndarray): Coulomb matrix of shape (...,N,N)
        unit_conversion (float) : Whether to scale units for distance. Default is 1.0.
        exponent (float): Exponent for diagonal elements. Default is 2.4.

    Returns:
        tuple: [inv_dist, z]

            - inv_dist (np.ndarray): Inverse distance Matrix of shape (...,N,N).
            - z (np.ndarray): Atom Number corresponding diagonal as proton number (..., N).
    """
    indslie = np.arange(0, coulomb_mat.shape[-1])
    z = coulomb_mat[..., indslie, indslie]
    z = np.power(2 * z, 1 / exponent)
    a = np.expand_dims(z, axis=len(z.shape) - 1)
    b = np.expand_dims(z, axis=len(z.shape))
    zz = a * b
    c = coulomb_mat / zz
    c[..., indslie, indslie] = 0
    c /= unit_conversion
    z = np.array(np.round(z), dtype=np.int)
    return c, z


def make_rotation_matrix(vector: np.ndarray, angle: float):
    r"""Generate rotation matrix around a given vector with a certain angle.

    Only defined for 3 dimensions explicitly here.

    Args:
        vector (np.ndarray, list): vector of rotation axis (3, ) with (x, y, z).
        angle (value): angle in degrees ° to rotate around.

    Returns:
        np.ndarray: Rotation matrix :math:`R` of shape (3, 3) that performs the rotation for :math:`y = R x`.
    """
    angle = angle / 180.0 * np.pi
    norm = (vector[0] ** 2.0 + vector[1] ** 2.0 + vector[2] ** 2.0) ** 0.5
    direction = vector / norm
    matrix = np.zeros((3, 3))
    matrix[0][0] = direction[0] ** 2.0 * (1.0 - np.cos(angle)) + np.cos(angle)
    matrix[1][1] = direction[1] ** 2.0 * (1.0 - np.cos(angle)) + np.cos(angle)
    matrix[2][2] = direction[2] ** 2.0 * (1.0 - np.cos(angle)) + np.cos(angle)
    matrix[0][1] = direction[0] * direction[1] * (1.0 - np.cos(angle)) - direction[2] * np.sin(angle)
    matrix[1][0] = direction[0] * direction[1] * (1.0 - np.cos(angle)) + direction[2] * np.sin(angle)
    matrix[0][2] = direction[0] * direction[2] * (1.0 - np.cos(angle)) + direction[1] * np.sin(angle)
    matrix[2][0] = direction[0] * direction[2] * (1.0 - np.cos(angle)) - direction[1] * np.sin(angle)
    matrix[1][2] = direction[1] * direction[2] * (1.0 - np.cos(angle)) - direction[0] * np.sin(angle)
    matrix[2][1] = direction[1] * direction[2] * (1.0 - np.cos(angle)) + direction[0] * np.sin(angle)
    return matrix


def rotate_to_principle_axis(coord: np.ndarray):
    r"""Rotate a point-cloud to its principle axis.

    This can be a molecule but also some general data.
    It uses PCA via SVD from :obj:`numpy.linalg.svd`. PCA from scikit uses SVD too (:obj:`scipy.sparse.linalg`).

    .. note::
        The data is centered before SVD but shifted back at the output.

    Args:
        coord (np.array): Array of points forming a pointcloud. Important: coord has shape (N,p)
            where N is the number of samples and p is the feature/coordinate dimension e.g. 3 for x,y,z

    Returns:
        tuple: [R, rotated]

            - R (np.array): Rotation matrix of shape (p, p) if input has (N,p)
            - rotated (np.array): Rotated point-could of coord that was the input.
    """
    centroid_c = np.mean(coord, axis=0)
    sm = coord - centroid_c
    zzt = (np.dot(sm.T, sm))  # Calculate covariance matrix
    u, s, vh = np.linalg.svd(zzt)
    # Alternatively SVD of coord with onyly compute vh but not possible for numpy/scipy.
    rotated = np.dot(sm, vh.T)
    rot_shift = rotated + centroid_c
    return vh, rot_shift


def rigid_transform(a: np.ndarray, b: np.ndarray, correct_reflection: bool = False):
    r"""Rotate and shift point-cloud A to point-cloud B. This should implement Kabsch algorithm.
    May also work for input of shape `(...,N,3)` but is not tested.
    Explanation of Kabsch Algorithm: https://en.wikipedia.org/wiki/Kabsch_algorithm
    For further literature:
    https://link.springer.com/article/10.1007/s10015-016-0265-x
    https://link.springer.com/article/10.1007%2Fs001380050048


    .. note::
        The numbering of points of A and B must match; not for shuffled point-cloud.
        This works for 3 dimensions only. Uses SVD.

    Args:
        a (np.ndarray): list of points (N,3) to rotate (and translate)
        b (np.ndarray): list of points (N,3) to rotate towards: A to B, where the coordinates (3) are (x,y,z)
        correct_reflection (bool): Whether to allow reflections or just rotations. Default is False.

    Returns:
        list: [A_rot, R, t]

            - A_rot (np.ndarray): Rotated and shifted version of A to match B
            - R (np.ndarray): Rotation matrix
            - t (np.ndarray): translation from A to B
    """
    a = np.transpose(np.array(a))
    b = np.transpose(np.array(b))
    centroid_a = np.mean(a, axis=1)
    centroid_b = np.mean(b, axis=1)
    am = a - np.expand_dims(centroid_a, axis=1)
    bm = b - np.expand_dims(centroid_b, axis=1)
    h = np.dot(am, np.transpose(bm))
    u, s, vt = np.linalg.svd(h)
    r = np.dot(vt.T, u.T)
    d = np.linalg.det(r)
    if d < 0:
        print("Warning: det(R)<0, det(R)=", d)
        if correct_reflection:
            print("Correcting R...")
            vt[-1, :] *= -1
            r = np.dot(vt.T, u.T)
    bout = np.dot(r, am) + np.expand_dims(centroid_b, axis=1)
    bout = np.transpose(bout)
    t = np.expand_dims(centroid_b - np.dot(r, centroid_a), axis=0)
    t = t.T
    return bout, r, t


def coordinates_from_distance_matrix(distance: np.ndarray, use_center: bool = None, dim: int = 3):
    r"""Compute list of coordinates from a distance matrix of shape `(N, N)`.
    May also work for input of shape `(..., N, N)` but is not tested.
    Uses vectorized Alogrithm:
    http://scripts.iucr.org/cgi-bin/paper?S0567739478000522
    https://www.researchgate.net/publication/252396528_Stable_calculation_of_coordinates_from_distance_information
    no check of positive semi-definite or possible k-dim >= 3 is done here
    performs svd from numpy

    Args:
        distance (np.ndarray): distance matrix of shape (N,N) with Dij = abs(ri-rj)
        use_center (int): which atom should be the center, dafault = None means center of mass
        dim (int): the dimension of embedding, 3 is default

    Return:
        np.ndarray: List of Atom coordinates [[x_1,x_2,x_3],[x_1,x_2,x_3],...]
    """
    distance = np.array(distance)
    dim_in = distance.shape[-1]
    if use_center is None:
        # Take Center of mass (slightly changed for vectorization assuming d_ii = 0)
        di2 = np.square(distance)
        di02 = 1 / 2 / dim_in / dim_in * (2 * dim_in * np.sum(di2, axis=-1) - np.sum(np.sum(di2, axis=-1), axis=-1))
        mat_m = (np.expand_dims(di02, axis=-2) + np.expand_dims(di02, axis=-1) - di2) / 2  # broadcasting
    else:
        di2 = np.square(distance)
        mat_m = (np.expand_dims(di2[..., use_center], axis=-2) + np.expand_dims(di2[..., use_center],
                                                                                axis=-1) - di2) / 2
    u, s, v = np.linalg.svd(mat_m)
    vec = np.matmul(u, np.sqrt(np.diag(s)))  # EV are sorted by default
    dist_out = vec[..., 0:dim]
    return dist_out


def get_principal_moments_of_inertia(masses: np.ndarray, coordinates: np.ndarray, shift_center_of_mass: bool = True):
    """Compute principle moments of inertia for a point-cloud with given mass.

    Args:
        masses (np.ndarray): Array of mass values.
        coordinates (np.ndarray): Coordinates of point-cloud.
        shift_center_of_mass (bool): Whether to shift to center of mass for inertia matrix.

    Returns:
        np.ndarray: Eigenvalues of inertia matrix. Shape is (3, ).
    """

    def translate_to_center_of_mass(m, xyz):
        if len(m.shape) <= 1:
            m = np.expand_dims(m, axis=-1)
        com = (np.sum(m * xyz, axis=0, keepdims=True) / np.sum(m))
        return xyz - com

    def get_inertia_matrix(m, xyz):
        x, y, z = xyz.T
        Ixx = np.sum(m * (y ** 2 + z ** 2))
        Iyy = np.sum(m * (x ** 2 + z ** 2))
        Izz = np.sum(m * (x ** 2 + y ** 2))
        Ixy = -np.sum(m * x * y)
        Iyz = -np.sum(m * y * z)
        Ixz = -np.sum(m * x * z)
        return np.array([[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]])

    coordinates = translate_to_center_of_mass(masses, coordinates) if shift_center_of_mass else coordinates
    inertia_matrix = get_inertia_matrix(masses, coordinates)
    pc = np.linalg.eigvals(inertia_matrix)
    pc = np.sort(pc)

    return pc


def shift_coordinates_to_unit_cell(coordinates: np.ndarray, lattice: np.ndarray):
    """Shift a set of coordinates into the unit cell of a periodic system.

    Args:
        coordinates (np.ndarray): Coordinate of nodes in the central primitive unit cell.
        lattice (np.ndarray): Lattice matrix of real space lattice vectors of shape `(3, 3)` .
            The lattice vectors must be given in rows of the matrix!

    Returns:
        np.ndarray: Coordinates shifted into unit cell defined by lattice.
    """
    lattice_inv = np.linalg.inv(lattice)
    frac_coordinates = np.dot(coordinates, lattice_inv)
    shifted_frac_coordinates = frac_coordinates % 1.0
    shifted_coordinates = np.dot(shifted_frac_coordinates, lattice)
    return shifted_coordinates


def distance_for_range_indices(coordinates: np.ndarray, indices: np.ndarray,
                               require_distance_dimension: bool = True):
    """Simple helper function to compute distances for indices.

    Args:
        coordinates (np.ndarray): Positions of shape `(N, 3)` .
        indices (np.ndarray): Pair-wise indices of shape `(M, 2)` .
        require_distance_dimension (bool): Whether to return distances of shape `(M, 1)` .

    Returns:
        np.ndarray: Distances of shape `(M, )` or `(M, 1)` if `require_distance_dimension` .
    """
    pos12 = coordinates[indices]
    dist = np.sqrt(np.sum(np.square(pos12[:, 0] - pos12[:, 1]), axis=-1))
    if require_distance_dimension:
        if len(dist.shape) <= 1:
            dist = np.expand_dims(dist, axis=-1)
    return dist


def distance_for_range_indices_periodic(coordinates: np.ndarray,
                                        indices: np.ndarray,
                                        images: np.ndarray,
                                        lattice: np.ndarray,
                                        require_distance_dimension: bool = True):
    """Simple helper function to compute distances for indices for a periodic system.

    Args:
        coordinates (np.ndarray): Positions of shape `(N, 3)` within the unit-cell.
        indices (np.ndarray): Pair-wise indices of shape `(M, 2)` .
        images (np.ndarray): Image translation of the second index of shape `(M, 3)` .
        lattice (np.ndarray): Lattice matrix of real space lattice vectors of shape `(3, 3)` .
            The lattice vectors must be given in rows of the matrix!
        require_distance_dimension (bool): Whether to return distances of shape `(M, 1)` .

    Returns:
        np.ndarray: Distances of shape `(M, )` or `(M, 1)` if `require_distance_dimension` .
    """
    pos12 = coordinates[indices]
    pos1, pos2 = pos12[:, 0], pos12[:, 1]
    if len(images.shape) > 2:
        # If image information are for both indices of shape (M, 2, 3)
        pos1 = pos1 + np.dot(images[:, 0], lattice)
        pos2 = pos2 + np.dot(images[:, 1], lattice)
    else:
        # Default is only for second index. Usually sending node.
        pos2 = pos2 + np.dot(images, lattice)
    dist = np.sqrt(np.sum(np.square(pos1 - pos2), axis=-1))
    if require_distance_dimension:
        if len(dist.shape) <= 1:
            dist = np.expand_dims(dist, axis=-1)
    return dist




from hashlib import md5
# import logging
import pymatgen.core.structure
from pymatgen.core.structure import Structure
from typing import Callable, Union
from networkx import MultiDiGraph
from kgcnn.graph.base import GraphDict

# A separate module logger is not need for the base class.
# logging.basicConfig()  # Module logger
# module_logger = logging.getLogger(__name__)
# module_logger.setLevel(logging.INFO)


class CrystalPreprocessor(Callable[[Structure], MultiDiGraph]):
    """Base class for crystal preprocessors.

    Concrete CrystalPreprocessors must be implemented as subclasses.
    """

    node_attributes = []
    edge_attributes = []
    graph_attributes = []

    def __init__(self, output_graph_as_dict: bool = False,
                 lattice: str = "graph_lattice", species: str = "node_number",
                 coords: str = "node_coordinates", charge: str = "charge"):
        self.output_graph_as_dict = output_graph_as_dict
        self._input_config = {
            "lattice": lattice, "species": species, "charge": charge, "coords": coords}

    def call(self, structure: Structure) -> MultiDiGraph:
        r"""Should be implemented in a subclass.

        Args:
            structure (Structure): Crystal for which the graph representation should be calculated.

        Raises:
            NotImplementedError:Should be implemented in a subclass.

        Returns:
            MultiDiGraph: Graph representation of the crystal.
        """
        raise NotImplementedError("Must be implemented in sub-classes.")

    def __call__(self, structure: Union[Structure, GraphDict]) -> Union[MultiDiGraph, GraphDict]:
        r"""Function to process crystal structures. Executes :obj:`call` .

        Args:
            structure (Structure): Crystal for which the graph representation should be calculated.

        Raises:
            NotImplementedError:Should be implemented in a subclass.

        Returns:
            MultiDiGraph: Graph representation of the crystal.
        """
        if isinstance(structure, GraphDict):
            structure = pymatgen.core.structure.Structure(
                lattice=structure.get(self._input_config["lattice"]),
                species=structure.get(self._input_config["species"]),
                coords=structure.get(self._input_config["coords"]),
                charge=structure.get(self._input_config["charge"]),
                coords_are_cartesian=True
            )
        nxg = self.call(structure)
        if self.output_graph_as_dict:
            g = GraphDict()
            g.from_networkx(
                nxg, node_attributes=self.node_attributes, edge_attributes=self.edge_attributes,
                graph_attributes=self.graph_attributes, reverse_edge_indices=True)
            return g
        return nxg

    def get_config(self) -> dict:
        """Returns a dictionary uniquely identifying the CrystalPreprocessor and its configuration.

        Returns:
            dict: A dictionary uniquely identifying the CrystalPreprocessor and its configuration.
        """
        config = vars(self)
        config = {k: v for k, v in config.items() if not k.startswith("_")}
        config['preprocessor'] = self.__class__.__name__
        config.update(self._input_config)
        return config

    def hash(self) -> str:
        """Generates a unique hash for the CrystalPreprocessor and its configuration.

        Returns:
            str: A unique hash for the CrystalPreprocessor and its configuration.
        """
        return md5(str(self.get_config()).encode()).hexdigest()

    def __hash__(self):
        return int(self.hash(), 16)

    def __eq__(self, other):
        return hash(self) == hash(other)



from kgcnn.layers.geom import NodePosition, NodeDistanceEuclidean, BesselBasisLayer, EdgeAngle, SphericalBasisLayer
from keras.layers import Concatenate, Subtract, Add
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.literature.DimeNetPP._layers import EmbeddingDimeBlock
from ._layers import MXMGlobalMP, MXMLocalMP


def model_disjoint(
    inputs,
    use_node_embedding,
    use_edge_embedding,
    input_node_embedding=None,
    input_edge_embedding=None,
    bessel_basis_local=None,
    spherical_basis_local=None,
    bessel_basis_global=None,
    use_edge_attributes=None,
    mlp_rbf_kwargs=None,
    mlp_sbf_kwargs=None,
    depth=None,
    global_mp_kwargs=None,
    local_mp_kwargs=None,
    node_pooling_args=None,
    output_embedding=None,
    use_output_mlp=None,
    output_mlp=None
):
    # Make input
    n, x, ed, ei_l, ri_g, ai_1, ai_2 = inputs[:7]
    batch_id_node, batch_id_edge, batch_id_ranges, batch_id_angles_1, batch_id_angles_2 = inputs[7:12]
    node_id, edge_id, range_id, angle_id1, angle_id2 = inputs[12:17]
    count_nodes, count_edges, count_ranges, count_angles1, count_angles2 = inputs[17:]

    # Rename to short names and make embedding, if no feature dimension.
    if use_node_embedding:
        n = EmbeddingDimeBlock(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = EmbeddingDimeBlock(**input_edge_embedding)(ed)

    # Calculate distances and spherical and bessel basis for local edges including angles.
    # For the first version, we restrict ourselves to 2-hop angles.
    pos1_l, pos2_l = NodePosition()([x, ei_l])
    d_l = NodeDistanceEuclidean()([pos1_l, pos2_l])
    rbf_l = BesselBasisLayer(**bessel_basis_local)(d_l)
    v12_l = Subtract()([pos1_l, pos2_l])
    a_l_1 = EdgeAngle()([v12_l, ai_1])
    a_l_2 = EdgeAngle(vector_scale=[1.0, -1.0])([v12_l, ai_2])
    sbf_l_1 = SphericalBasisLayer(**spherical_basis_local)([d_l, a_l_1, ai_1])
    sbf_l_2 = SphericalBasisLayer(**spherical_basis_local)([d_l, a_l_2, ai_2])

    # Calculate distance and bessel basis for global (range) edges.
    pos1_g, pos2_g = NodePosition()([x, ri_g])
    d_g = NodeDistanceEuclidean()([pos1_g, pos2_g])
    rbf_g = BesselBasisLayer(**bessel_basis_global)(d_g)

    if use_edge_attributes:
        rbf_l = Concatenate()([rbf_l, ed])

    rbf_l = GraphMLP(**mlp_rbf_kwargs)([rbf_l, batch_id_edge, count_edges])
    sbf_l_1 = GraphMLP(**mlp_sbf_kwargs)([sbf_l_1, batch_id_angles_1, count_angles1])
    sbf_l_2 = GraphMLP(**mlp_sbf_kwargs)([sbf_l_2, batch_id_angles_2, count_angles2])
    rbf_g = GraphMLP(**mlp_rbf_kwargs)([rbf_g, batch_id_ranges, count_ranges])

    # Model
    h = n
    nodes_list = []
    for i in range(0, depth):
        h = MXMGlobalMP(**global_mp_kwargs)([h, rbf_g, ri_g])
        h, t = MXMLocalMP(**local_mp_kwargs)([h, rbf_l, sbf_l_1, sbf_l_2, ei_l, ai_1, ai_2])
        nodes_list.append(t)

    # Output embedding choice
    out = Add()(nodes_list)
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)([count_nodes, out, batch_id_node])
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `MXMNet`")

    return out



import keras as ks
from keras.layers import Dense
from kgcnn.layers.conv import GIN, GINE
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.pooling import PoolingNodes


def model_disjoint(inputs,
                   use_node_embedding: bool = None,
                   input_node_embedding: dict = None,
                   depth: int = None,
                   gin_args: dict = None,
                   gin_mlp: dict = None,
                   last_mlp: dict = None,
                   dropout: float = None,
                   output_embedding: str = None,
                   output_mlp: dict = None):
    n, disjoint_indices, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)

    # Model
    # Map to the required number of units.
    n_units = gin_mlp["units"][-1] if isinstance(gin_mlp["units"], list) else int(gin_mlp["units"])
    n = Dense(n_units, use_bias=True, activation='linear')(n)
    list_embeddings = [n]
    for i in range(0, depth):
        n = GIN(**gin_args)([n, disjoint_indices])
        n = GraphMLP(**gin_mlp)([n, batch_id_node, count_nodes])
        list_embeddings.append(n)

    # Output embedding choice
    if output_embedding == "graph":
        out = [PoolingNodes()([count_nodes, x, batch_id_node]) for x in list_embeddings]  # will return tensor
        out = [MLP(**last_mlp)(x) for x in out]
        out = [ks.layers.Dropout(dropout)(x) for x in out]
        out = ks.layers.Add()(out)
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":  # Node labeling
        out = GraphMLP(**last_mlp)([n, batch_id_node, count_nodes])
        out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `GIN` .")

    return out


def model_disjoint_edge(
        inputs,
        use_node_embedding: bool = None,
        use_edge_embedding: bool = None,
        input_node_embedding: dict = None,
        input_edge_embedding: dict = None,
        depth: int = None,
        gin_args: dict = None,
        gin_mlp: dict = None,
        last_mlp: dict = None,
        dropout: float = None,
        output_embedding: str = None,
        output_mlp: dict = None):

    n, ed, disjoint_indices, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    # Model
    # Map to the required number of units.
    n_units = gin_mlp["units"][-1] if isinstance(gin_mlp["units"], list) else int(gin_mlp["units"])
    n = Dense(n_units, use_bias=True, activation='linear')(n)
    ed = Dense(n_units, use_bias=True, activation='linear')(ed)
    list_embeddings = [n]
    for i in range(0, depth):
        n = GINE(**gin_args)([n, disjoint_indices, ed])
        n = GraphMLP(**gin_mlp)([n, batch_id_node, count_nodes])
        list_embeddings.append(n)

    # Output embedding choice
    if output_embedding == "graph":
        out = [PoolingNodes()([count_nodes, x, batch_id_node]) for x in list_embeddings]  # will return tensor
        out = [MLP(**last_mlp)(x) for x in out]
        out = [ks.layers.Dropout(dropout)(x) for x in out]
        out = ks.layers.Add()(out)
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":  # Node labeling
        out = GraphMLP(**last_mlp)([n, batch_id_node, count_nodes])
        out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `GINE` .")

    return out


# import keras_core as ks
import functools
import logging
from math import inf
from typing import Union
from copy import deepcopy
import importlib


# Module logger
logging.basicConfig()
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


def get_model_class(module_name: str, class_name: str):
    r"""Helper function to get model class by string identifier.

    Args:
        module_name (str): Name of the module of the model.
        class_name (str): Name of the model class.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    if module_name[:6] != "kgcnn.":
        # Assume that is simply the module name in kgcnn.literature.
        module_name = "kgcnn.literature.%s" % module_name
    if class_name is None or class_name == "":
        # Assume that "make_model" function is used.
        class_name = "make_model"

    try:
        make_model = getattr(importlib.import_module(module_name), class_name)
    except ModuleNotFoundError:
        raise NotImplementedError("Unknown model identifier '%s' for a model in kgcnn.literature." % class_name)

    return make_model


def update_model_kwargs_logic(default_kwargs: dict = None, user_kwargs: dict = None,
                              update_recursive: Union[int, float] = inf):
    r"""Make model kwargs dictionary with updated default values. This is essentially a nested version of update()
    for dicts. This is supposed to be more convenient if the values of kwargs are again layer kwargs to be unpacked,
    and do not need to be fully known to update them.

    Args:
        default_kwargs (dict): Dictionary of default values. Default is None.
        user_kwargs (dict): Dictionary of args to update. Default is None.
        update_recursive (int): Max depth to update mappings like dict. Default is `inf`.

    Returns:
        dict: New dict and update with first default and then user args.
    """
    if default_kwargs is None:
        default_kwargs = {}
    if user_kwargs is None:
        user_kwargs = {}

    # Check valid kwargs
    for iter_key in user_kwargs.keys():
        if iter_key not in default_kwargs:
            raise ValueError("Model kwarg {0} not in default arguments {1}".format(iter_key, default_kwargs.keys()))

    # Start with default values.
    out = deepcopy(default_kwargs)

    # Nested update of kwargs:
    def _nested_update(dict1, dict2, max_depth=inf, depth=0):
        for key, values in dict2.items():
            if key not in dict1:
                module_logger.warning("Model kwargs: Unknown key {0} with value {1}".format(key, values))
                dict1[key] = values
                continue
            if not isinstance(dict1[key], dict):
                dict1[key] = values
                continue
            if not isinstance(values, dict):
                module_logger.warning("Model kwargs: Overwriting dictionary of {0} with {1}".format(key, values))
                dict1[key] = values
                continue
            # Nested update.
            if depth < max_depth:
                dict1[key] = _nested_update(dict1[key], values, max_depth=max_depth, depth=depth+1)
            else:
                dict1[key] = values
        return dict1

    return _nested_update(out, user_kwargs, update_recursive, 0)


def update_model_kwargs(model_default, update_recursive=inf, deprecated: list = None):
    """Decorating function for update_model_kwargs_logic() ."""
    def model_update_decorator(func):

        @functools.wraps(func)
        def update_wrapper(*args, **kwargs):

            updated_kwargs = update_model_kwargs_logic(model_default, kwargs, update_recursive)

            # Logging of updated values.
            if 'verbose' in updated_kwargs:
                module_logger.setLevel(updated_kwargs["verbose"])
            module_logger.info("Updated model kwargs: '%s'." % updated_kwargs)

            if len(args) > 0:
                module_logger.error("Can only update kwargs, not %s" % args)

            return func(*args, **updated_kwargs)

        return update_wrapper

    return model_update_decorator


def change_attributes_in_all_layers(model, attributes_to_change=None, layer_type=None):
    r"""Utility/helper function to change the attributes from a dictionary in all layers of a model of a certain type.

    .. warning::

        This function can change attributes but can cause problems for built models. Also take care which attributes
        you are changing, especially if they include weights. Always check model behaviour after applying this function.

    Args:
        model (tf.keras.models.Model): Model to modify.
        attributes_to_change (dict): Dictionary of attributes to change in all layers of a specific type.
        layer_type: Class type of the layer to change. Default is None.

    Returns:
        tf.keras.models.Model: Model which has layers with changed attributes.
    """
    if model.built:
        module_logger.warning("Model '%s' has already been built. Will set `built=False` and continue." % model.name)
        model.built = False
    if attributes_to_change is None:
        attributes_to_change = {}
    all_layers = model._flatten_layers(include_self=False, recursive=True)
    for x in all_layers:
        if layer_type is not None:
            if not isinstance(x, layer_type):
                continue
        changed_attributes = False
        for key, value in attributes_to_change.items():
            if hasattr(x, key):
                setattr(x, key, value)
                changed_attributes = True
        if changed_attributes:
            if x.built:
                module_logger.warning(
                    "Layer '%s' in model has already been built. Will set `built=False` and continue." % x.name)
                x.built = False
    return model
